(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontmodule.ml,v 1.2 2001/12/10 18:20:13 malekith Exp $
 *
 * Module implementation: Gontmodule
 * Description: We implement module interface here. As Gont modules are
 * very simple, so *this* module is :)
 *)

open Gontutil

type mv_record = {mv_name: id;
                  mv_type: ty;
                  mutable mv_symbol: symbol option}

type mt_record = {mt_name: id;
                  mt_tydef: typedef}
                  
type module_symbol =
     Value of mv_record
   | Type of mt_record

type mod_record = {mod_name: id;
                   mod_vals: mv_record list;
                   mod_types: mt_record list}

(* We use it as magic. It is stored in all .gio files, if file contain
 * bad indent string -- it is rejected. This is probably not enough 
 * (we should also put version of gontutil.mli here, but... nothing's
 * perfect :^) *)
let ver = "Gont Compiled Interface, $Revision: 1.2 $"

(* We often use "" as name of current module. This cannot be the case
 * for module representation that is going to be saved. *)
let set_module_name m t =
  let f t =
    match t with
        NamedType r -> NamedType {r with nt_module = m}
      (* Don't use `| _ -> t' here, so we'll get warned if sth
       * new is added. *)
      | Int | Void | Bool | Float | String | TyVar _ | AbstractType
      | Tuple _ | Func _ | Struct _ | OptStruct _ | Union _ -> t in
  type_walk f t
;;

let save m out_ch decls =
  let rec make_mv r decls =
    match decls with
        x::xs ->
          (match x with
              GlobalDecl {global_type = t; global_name = n}
            | FuncProto {fp_type = t; fp_name = n} ->
                make_mv  ({mv_name = n; 
                           mv_type = (set_module_name m t); 
                           mv_symbol = None} :: r) xs
            | TypeDef _ -> (* skip *) make_mv r xs
            | FuncDecl _
            | ExternC _
            | TopLevelOpen _ -> assert false)
      | [] -> r in
  let rec make_mt r decls =
    match decls with
        x::xs ->
          (match x with
              GlobalDecl _ | FuncProto _ -> (* skip *) make_mt r xs
            | TypeDef ({td_name = n; td_type = t} as tt) -> 
                make_mt ({mt_name = n; 
                          mt_tydef = {tt with 
                                      td_type = (set_module_name m t);
                                      td_module = m}} :: r) xs
            | FuncDecl _
            | ExternC _
            | TopLevelOpen _ -> assert false)
      | [] -> r in
  let r = {mod_name = m; 
           mod_vals = make_mv [] decls;
           mod_types = make_mt [] decls} in
  Marshal.to_channel out_ch ver [];
  Marshal.to_channel out_ch r []
;;

let modules : (string, mod_record) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear modules

exception Cannot_open
    
let open_module m =
  let load m =
    let rec open_any names =
      match names with
          x::xs -> 
            (try
              open_in x
             with 
               Sys_error _ -> open_any xs)
        | [] -> raise Cannot_open in
    let in_ch = 
      try
        open_any [((String.uncapitalize m) ^ ".gio")]
      with Cannot_open -> 
        raise (Error (xf "cannot find module interface for %s" m)) in
    let id = (Marshal.from_channel in_ch : string) in
      if id <> ver then
        (close_in in_ch; 
         raise (Error (xf "module interface file for %s was compiled \
                           with diffrent version, please recompile \
                           the interface" m)))
      else
        let mr = (Marshal.from_channel in_ch : mod_record) in
          if mr.mod_name <> m then
            (close_in in_ch;
             raise (Error (xf "interface file contains module %s, \
                               instead of expected %s" mr.mod_name m)))
           else
             Hashtbl.add modules mr.mod_name mr;
             close_in in_ch in
  try
    ignore(Hashtbl.find modules m)
  with Not_found -> load m
;;

let find_module m =
  open_module m;
  Hashtbl.find modules m

exception Has_no_symbol of ty
exception Symbol_found of symbol
exception Tydef_found of typedef

let import_symbol mod_name sym_name =
  try
    List.iter (fun s -> 
       match s with
          {mv_name = n; mv_symbol = Some sym} when (n = sym_name) -> 
            raise (Symbol_found sym)
        | {mv_name = n; mv_type = t} when (n = sym_name) -> 
            raise (Has_no_symbol t)
        | _ -> ()) (find_module mod_name).mod_vals;
    raise Not_found
  with Symbol_found s -> s
  
let import_type_symbol mod_name sym_name =
  try
    List.iter (fun s -> 
       match s with
          {mt_name = n; mt_tydef = t} when (n = sym_name) -> 
            raise (Tydef_found t)
        | _ -> ()) (find_module mod_name).mod_types;
    raise Not_found
  with Tydef_found t -> t
    
let assign_symbol mod_name sym_name sym =
  let m = find_module mod_name in
  let rec find l =
    match l with
        {mv_name = n; mv_symbol = Some _}::_ when (n = sym_name) -> 
          assert false
      | ({mv_name = n} as r)::_ when (n = sym_name) -> 
          r.mv_symbol <- Some sym
      | _::xs -> 
          find xs
      | [] -> 
          raise Not_found in
   find m.mod_vals

let globals mod_name =
  let m = find_module mod_name in
  ((List.map (fun {mv_name = n; mv_type = t} -> (t, n)) m.mod_vals),
   (List.map (fun {mt_name = n; mt_tydef = t} -> t) m.mod_types))
;;
